/* 
 * Copyright (C) 2013 Luca Zanetti. All rights reserved.
 *
 * This file may be used under the terms of the GNU General Public License
 * versions 3.0 as published by the Free Software Foundation and
 * appearing in the COPYING file included in the packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 *  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "qmc.h"
#include "list.h"

/* read and return the adjacency matrix of a n vertex graph from file */
Graph read_graph_f(const char *filename, int n) {
	Graph g = calloc(n, sizeof(int*));
	int i,j;
	for(i=0; i<n; i++) {
		g[i] = calloc(n, sizeof(int));
	}
	FILE *fp = fopen(filename,"r");
	for(i=0; i<n; i++) {
		for (j=0; j<n; j++) {
			fscanf(fp, " %d", &g[i][j]);
		}
	}
	fclose(fp);
	return g;
}

/* print the current configuration of replica r */
void print_state_r(State state, int r) {
	int i;
	for (i=0; i<state->n; i++) {
		printf("%d ", state->replicas[r][i]);
	}
}

/* print the current configuration of chain c in every replica */
void print_state_c(State state, int c) {
	int i;
	for (i=0; i<2*M; i++) {
		printf("%d: %d -> %d\n", i, c, state->replicas[i][c]);
	}
}

/* print the current configuration of every replica */
void print_state_a(State state, Graph g1, Graph g2) {
	int r;
	for (r=0; r<2*M; r++) {
		printf("Replica %d:  ", r);
		print_state_r(state, r);
		printf("  Cost: %d\n", configuration_cost(state, g1, g2, r));
	}
	printf("\n\n\n");
}

/*
double coeff_weight(int a, int b, int c, int d, double gamma) {
	if (a==b==c==d)
		return 1;
	if (a==c==0 && b==d==1)
		return cosh(gamma*BETA/M);
	if (a==c==1 && b==d==0)
		return cosh(gamma*BETA/M);
	if (a+b==1 && c+d==1)
		return sinh(gamma*BETA/M);
}*/

/* calculate the local weight of chain c of replica r (from potential energy)*/
double diagonal_weight(State state, Graph g1, Graph g2, int c, int r, double beta, double gamma) {
	int i, j, n_i;
	int w = 0;
	/* if state of chain c is j, count # of i s.t. 
	 	{c,i} in g1, state of chain i is k and {j,k} not in g2 */	
	for (i=0; i<state->n; i++) {
		if (g1[c][i]) {
			//~ if (r==2*M-1) printf("state->replicas[%d][%d] = %d\n",r,c,state->replicas[r][c]);
			if(!g2[state->replicas[r][c]][state->replicas[r][i]])
				w++;
		} else {
			if(g2[state->replicas[r][c]][state->replicas[r][i]])
				w++;
		}
	}
	//~ /* if state of chain c is j, count # of i s.t. 
	 	//~ {c,i} in g2, state of chain i is k and {j,k} not in g1 */
	//~ for (i=0; i<state->n; i++) {
		//~ if (g2[c][i]) 
			//~ if(!g1[state->replicas[r][c]][state->replicas[r][i]])
				//~ w++;
	//~ }
	//~ 
	/* non-permutation cost */
	/*for (i=0; i<state->n; i++) {
		if (i != c && state->replicas[r][c] == state->replicas[r][i])
			w ++;
	}*/
	for (i=0; i<state->n; i++) 
		for (j=i+1; j<state->n; j++) 
			if (state->replicas[r][i] == state->replicas[r][j])
				w ++;
	
	#ifdef ADIABATIC
	return exp(-w*(1-gamma)*beta/M);
	#else
	return exp(-w*beta/M);
	#endif
}

/* sample a valid plaquette from the current configuration of chain c. Returns the replica number.
 	The seed for the random number generator must be initialized */
int sample_plaquette(State state, int c) {
	List valid_plaquettes = NULL;
	int r, p;
	int n_v = 0;
	for (r=0; r<2*M-1; r++) {
		if (state->replicas[r][c] == state->replicas[r+1][c]) {
			if ((r%2 == 0 && state->replicas[r][c] == (state->n-1) && (state->n-1)%2 == 0) 
				|| (r%2 != 0 && state->replicas[r][c] == 0)
				|| (r%2 != 0 && state->replicas[r][c] == (state->n-1) && (state->n-1)%2 != 0))
						continue; /* check plaquette at the boundary */
			valid_plaquettes = add(valid_plaquettes, r);
			n_v++;
		}
	}
	if (state->replicas[2*M-1][c] == state->replicas[0][c] && state->replicas[2*M-1][c] != 0 &&
			!(state->replicas[2*M-1][c] == state->n-1 && (state->n-1)%2 != 0)) {
		valid_plaquettes = add(valid_plaquettes, 2*M-1);
		n_v++;
	}
	int r_number = rand();
	//~ printf("\nRRRR  %d\n", r_number);
	p = val(valid_plaquettes, r_number%n_v);
	destroy(valid_plaquettes);
	return p;
}

/* compute local weight of chain c of replica r (from kinetic energy):
	chain c of replica r must be in the same state as chain c of replica r+1 to be a valid plaquette */
double kinetic_weight(State state, int c, int r, double gamma, double beta) {
	int next_r = (r+1) % (2*M);
	int next_next_r = (r+2) % (2*M);
	int prev_r = (r+2*M-1) % (2*M);
	double w = 1.0;
	/* calculate the weight of the plaquette */
	if (state->replicas[next_r][c] == state->replicas[next_next_r][c])
		w *= cosh(gamma*beta/M);
	else
		w *= sinh(gamma*beta/M);
	if (state->replicas[r][c] == state->replicas[prev_r][c])
		w *= cosh(gamma*beta/M);
	else
		w *= sinh(gamma*beta/M);
	return w;
}

/* modify the local configuration of the plaquette of chain c and replicas (r,r+1)
 * return 0 if the move has been correctly executed */
int plaquette_move(State state, int c, int r) {
	int next_r = (r+1) % (2*M);
	if (state->replicas[r][c] != state->replicas[next_r][c])
		return -1; /* not a valid plaquette */
	if (r%2 == 0) { 
		if (state->replicas[r][c]%2 == 0) {
			state->replicas[r][c]++;
			state->replicas[next_r][c]++;
		} else {
			state->replicas[r][c]--;
			state->replicas[next_r][c]--;
		}
	} else {
		if (state->replicas[r][c]%2 != 0) {
			state->replicas[r][c]++;
			state->replicas[next_r][c]++;
		} else {
			state->replicas[r][c]--;
			state->replicas[next_r][c]--;
		}
	}
	return 0;
}

/* make a move according to metropolis rule:
 * choose u.a.r a chain c and a valid plaquette r
 * compute current weight w of the plaquette
 * flip the spins of the plaquette (modifying chain c of replicas r and r+1)
 * compute new weight w'
 * accept the move with probability min{1,w'/w} and return 0 otherwise reject it and return 1*/
int metropolis_move_old(State state, Graph g1, Graph g2, double gamma, double beta) {
	int r_number = rand();
	//~ printf("\nRRRR  %d\n", r_number);
	int c = r_number % state->n;
	int r = sample_plaquette(state, c);
	/*printf("r=%d,c=%d\n",r,c);*/
	int next_r = (r+1) % (2*M);
	double w_curr, w_new;
	w_curr = diagonal_weight(state, g1, g2, c, r, beta, gamma) * diagonal_weight(state, g1, g2, c, next_r, beta, gamma);
	w_curr *= kinetic_weight(state, c, r, gamma, beta);
	/*printf("w_curr=%.3f\n",w_curr);*/
	plaquette_move(state, c, r);
	/*printf("r=%d c=%d\t",r,c);*/
	w_new = diagonal_weight(state, g1, g2, c, r, beta, gamma) * diagonal_weight(state, g1, g2, c, next_r, beta, gamma);
	w_new *= kinetic_weight(state, c, r, gamma, beta);
	if (w_curr <= w_new) {
		acc++;
		return 0; /* accept the move: stay in this configuration */
	}
	double p = w_new / w_curr;
	r_number = rand();
	//~ printf("\nRRRR  %d\n", r_number);
	if (((double)r_number/(double)RAND_MAX) < p) {
		acc++;
		return 0; /* accept the move */
	} else {
		ref++;
		plaquette_move(state, c, r); /* reject the move: return to previous configuration, else accept */
		return -1;
	}
}


/* make a move according to metropolis rule:
 * (for every chain) choose u.a.r a valid plaquette r
 * compute current weight w of the plaquette
 * flip the spins of the plaquette (modifying chain c of replicas r and r+1)
 * compute new weight w'
 * accept the move with probability min{1,w'/w} */
void metropolis_move(State state, Graph g1, Graph g2, double gamma, double beta) {
	State old_state = malloc(sizeof(struct state));
	int r,c;
	old_state->n = state->n;
	old_state->replicas = malloc((2*M)*sizeof(int *));
	for (r=0; r<2*M; r++) {
		old_state->replicas[r] = malloc((state->n)*sizeof(int*));
		for (c=0; c<state->n; c++) {
			old_state->replicas[r][c] = state->replicas[r][c];
		}
	}
	
	for (c=0; c<state->n; c++) {
		int r = sample_plaquette(old_state, c);
		//~ printf("r=%d,c=%d\n",r,c);
		int next_r = (r+1) % (2*M);
		double w_curr, w_new;
		w_curr = diagonal_weight(old_state, g1, g2, c, r, beta, gamma) * diagonal_weight(old_state, g1, g2, c, next_r, beta, gamma);
		//~ printf("w_curr_diag=%.6f\n",w_curr);
		w_curr *= kinetic_weight(old_state, c, r, gamma, beta);
		//~ printf("w_curr=%.6f\n",w_curr);
		plaquette_move(old_state, c, r);
		/*printf("r=%d c=%d\t",r,c);*/
		w_new = diagonal_weight(old_state, g1, g2, c, r, beta, gamma) * diagonal_weight(old_state, g1, g2, c, next_r, beta, gamma);
		//~ printf("w_new_diag=%.6f\n",w_new);
		w_new *= kinetic_weight(old_state, c, r, gamma, beta);
		//~ printf("p=%.6f\n",w_new/w_curr);
		if (w_curr <= w_new) { /* accept the move: stay in this configuration */
			plaquette_move(state, c, r);
			plaquette_move(old_state, c, r); 
			acc++;
		} else {	
			double p = w_new / w_curr;
			int r_number = rand();
			//~ printf("\nRRRR  %d\n", r_number);
			if (((double)r_number/(double)RAND_MAX) < p) { /* accept the move */
				plaquette_move(state, c, r);
				plaquette_move(old_state, c, r); 
				acc++;
			} else { /* reject the move: return to previous configuration, else accept */
				plaquette_move(old_state, c, r); 
				ref++;
			}
		}	
	}
	for (r=0; r<2*M; r++) {
		free(old_state->replicas[r]);
	}
	free(old_state->replicas);
	free(old_state);
}


/* create and initialize a new state of the system: 
 * every replica is initialized as the identity permutation (-> vertical world-lines)
 * */
State init(Graph g1, Graph g2, int n) {
	State state = malloc(sizeof(struct state));
	state->n = n;
	state->replicas = malloc((2*M)*sizeof(int *));
	int r,c;
	/*for (r=0; r<2*M; r++) {
		state->replicas[r] = malloc(n*sizeof(int*));
		for (c=0; c<n; c++) {
			state->replicas[r][c] = c;
			//~ state->replicas[r][c] = 1;
			//~ state->replicas[r][c] = n/2;
		}
	}*/
	r = 0;
	state->replicas[r] = malloc(n*sizeof(int*));
	for (c=0; c<n; c++) {
		state->replicas[r][c] = rand() % n;
	}
	for (r=1; r<2*M; r++) {
		state->replicas[r] = malloc(n*sizeof(int*));
		for (c=0; c<n; c++) {
			state->replicas[r][c] = state->replicas[0][c];
		}

	}
	return state;
}

/* return cost of the configuration currently assumed by replica r */
int configuration_cost(State state, Graph g1, Graph g2, int r) {
	int i,c,j,n_i;
	int w = 0;
	for (c=0; c<state->n; c++) {
		/* if state of chain c is j, count # of i s.t. 
		 	{c,i} in g1, state of chain i is k and {j,k} not in g2 */	
		for (i=0; i<state->n; i++) {
			if (g1[c][i] == 1) {
				/*printf("state->replicas[%d][%d] = %d\n",r,c,state->replicas[r][c]);*/
				if(g2[state->replicas[r][c]][state->replicas[r][i]] == 0)
					w++;
			} else {
				if(g2[state->replicas[r][c]][state->replicas[r][i]] == 1)
					w++;
			}
		}
	}
	/*printf("w=%d\n",w);*/
	/* non-permutation cost */
	for (i=0; i<state->n; i++) {
		n_i = 0;
		for (j=0; j<state->n; j++) {
			if (state->replicas[r][j] == i) n_i++;
		}
		w += (n_i - 1)*(n_i - 1);
	}

	return w;
}

#ifndef DWAVE

#ifdef SIMULATION
int main(int argc, char *argv[]) {
	#ifndef PREANNEAL
	int n = atoi(argv[1]);
	int t = atoi(argv[2]);
	double gamma_init = atof(argv[3]);
	double gamma_min = atof(argv[4]);
	double gamma_coeff = atof(argv[5]);
	
	double beta_init = 1;
	double beta_coeff = 1.01;
	double beta = beta_init;
	
	srand(time(NULL));
	Graph g1 = read_graph_f(INPUT1, n);
	Graph g2 = read_graph_f(INPUT2, n);
	State state = init(g1, g2, n);
	
	
	int k,l,i,j;
	double gamma = gamma_init;
	while (gamma >= gamma_min) {
		beta = beta_init;
		for (l=0; l<300; l++) {
			for (k=0; k<t; k++) {
				metropolis_move(state, g1, g2, gamma, beta);
			}
			beta *= beta_coeff;
		}
		printf("\n\ngamma=%.12f\n", gamma);
		/*for (i=0; i<2*M-1;i++) {
			printf("R%d:  ",i);
			for (j=0; j<n-1;j++) {
				printf("%d,",state->replicas[i][j]);
			}
			printf("%d\n",state->replicas[i][j]);
		}*/
		print_state_a(state, g1, g2);
		printf("\n***");
		gamma *= gamma_coeff;
	}
	#else
	int n = atoi(argv[1]);
	int k = atoi(argv[2]);
	int tmax = atoi(argv[3]);
	M = atoi(argv[4]);
	
	double beta;
	
	srand(time(NULL));
	Graph g1 = read_graph_f(INPUT1, n);
	Graph g2 = read_graph_f(INPUT2, n);
	State state = init(g1, g2, n);
	double t=1;
	int i;
	double gamma = 1;
	while (t <= 5*M) {
		beta = M*t/100;
		for (i=0; i<k;i++)
			metropolis_move(state, g1, g2, gamma, beta);
		t*=1.05;
	}
	#ifdef VERBOSE
	printf("gamma=%.8f\n", gamma);
	print_state_a(state, g1, g2);
	printf("\n***");
	#endif
	t=tmax;
	beta = M;
	while (t > 0) {
		gamma = t/tmax;
		for (i=0; i<k;i++)
			metropolis_move(state, g1, g2, gamma, beta);
		#ifdef VERBOSE
		printf("gamma=%.8f\n", gamma);
		print_state_a(state, g1, g2);
		printf("\n***");
		#endif
		t -= 1;
	}
	#endif
	#ifdef VERBOSE
	printf("gamma=%.8f\n", gamma);
	print_state_a(state, g1, g2);
	int s = 0;
	for (i=0; i<2*M; i++) {
		if (configuration_cost(state, g1, g2, i) == 0)
			s++;
	}
	if (s >= M/3) {
		printf("\n\nY %d", s);
	} else {
		printf("\n\nN %d", s);
	}
	#else
	print_state_a(state, g1, g2);
	int s = 0;
	for (i=0; i<2*M; i++) {
		if (configuration_cost(state, g1, g2, i) == 0)
			s++;
	}
	if (s >= M/3) {
		printf("\n\n%d", s);
	} else {
		printf("\n\n%d", s);
	}
	/*if (s >= M/3) {
		printf("Y");
	} else {
		printf("N");
	}*/
	#endif
	
	return 0;
}


#else
#ifdef FIXGAMMA
int main(int argc, char *argv[]) {
	int n = atoi(argv[1]);
	int t = atoi(argv[2]);
	double gamma = atof(argv[3]);
	double beta_init = atof(argv[4]);
	double beta_coeff = atof(argv[5]);
	double anneal_t = atoi(argv[6]);
	
	
	double beta = beta_init;
	srand(time(NULL));
	Graph g1 = read_graph_f(INPUT1, n);
	Graph g2 = read_graph_f(INPUT2, n);
	State state = init(g1, g2, n);
	
	printf("FIXGAMMA gamma=%.3f\n", gamma);
	
	int i,j;
	for (i=0; i<anneal_t; i++) {
		for (j=0; j<t; j++) {
			metropolis_move(state, g1, g2, gamma, beta);
		}
		#ifdef PRINT_A
		printf("\n\nbeta=%.12f\n", beta);
		for (i=0; i<2*M;i++) {
			printf("R%d:  ",i);
			for (j=0; j<n-1;j++) {
				printf("%d,",state->replicas[i][j]);
			}
			printf("%d\n",state->replicas[i][j]);
		}
		printf("\n***");
		#endif
		
		beta *= beta_coeff;
	}
	
	#ifdef CONV
	int conv = 0;
	while (conv == 0){
		conv = 1;
		for (j=0; j<t; j++) {
			metropolis_move(state, g1, g2, gamma, beta);
		}
		for (i=1; i<2*M; i++) {
			for (j=0; j<n; j++) {
				if (state->replicas[0][j] != state->replicas[i][j])
					conv = 0;
			}
		}
		if (conv == 1) {
			printf("R%d:  ",0);
			for (j=0; j<n-1;j++) {
				printf("%d,",state->replicas[0][j]);
			}
			printf("%d\n",state->replicas[0][j]);
		}
		beta *= beta_coeff;
	}
	#else
	for (i=0; i<2*M;i++) {
		printf("R%d:  ",i);
		for (j=0; j<n-1;j++) {
			printf("%d,",state->replicas[i][j]);
		}
		printf("%d\n",state->replicas[i][j]);
	}
	printf("\n***");
	#endif
	
	return 0;
}
#else
int main(int argc, char *argv[]) {
	int n = atoi(argv[1]);
	int t = atoi(argv[2]);
	int k,h,l,r,c,conv,min;
	
	double gamma_init = 25;
	double beta_init = 1;
	double beta_max = beta_init;
	double gamma = gamma_init;
	double beta = beta_init;
	srand(time(NULL));
	Graph g1 = read_graph_f(INPUT1, n);
	Graph g2 = read_graph_f(INPUT2, n);
	State state = init(g1, g2, n);
		
	int min_state[n];
	min = configuration_cost(state, g1, g2, 0);
	memcpy(min_state, state->replicas[0], sizeof(int[n]));
	
	printf("NORMAL\n");
	
	acc=0;
	ref=0;
	gamma = gamma_init;
	for (k=0; k<500; k++) {
		beta = beta_init;
		for (l=0; l<1000; l++) {
			beta_max = (beta > beta_max) ? beta : beta_max;
			for (h=0; h<t; h++) {
				metropolis_move(state, g1, g2, gamma, beta);
				for (r=0; r<2*M; r++) {
					if (configuration_cost(state, g1, g2, r) < min) {
						min = configuration_cost(state, g1, g2, r);
						memcpy(min_state, state->replicas[r], sizeof(int[n]));
					}
				}
			}
				//~ metropolis_move(state, g1, g2, gamma);
			beta *= 1.01;
		}
		gamma *= 0.95;
		//~ gamma -= (gamma*0.1);
		/*beta_init += beta_init * 0.01;*/
		printf("%d - gamma=%.12f\tacc=%d\tref=%d\n", k, gamma, acc, ref);
		
		if (conv == 2) continue;
		conv = 1;
		for (r=1; r<2*M; r++) {
			for (c=0; c<n; c++) {
				if (state->replicas[0][c] != state->replicas[r][c])
					conv = 0;
			}
		}
		if (conv == 1) {
			conv = 2;
			printf("converged in %d iterations!\n\n", k+1);
			print_state_a(state, g1, g2);
			printf("\n\n\n\n*****************************************************\n");
		}
	}
	print_state_a(state, g1, g2);
	
	/*static const int conf[6] = {5,4,2,1,0,3};
	//{6, 5, 3, 2, 1, 4}
	
	memcpy(state->replicas[0], conf, sizeof(conf));*/
	
	printf("min cost = %d\n", min );
	for (c=0; c<n; c++) {
		printf("%d -> %d\n", c, min_state[c]);
	}
	printf("max beta = %.12f\n", beta_max);
	
	return 0;
}
#endif
#endif


#endif
